@@ -24,7 +24,7 @@
  * GTK+ at ftp://ftp.gtk.org/pub/gtk/. 
  */
 
-#include <config.h>
+#include "config.h"
 #include <math.h>
 #include <pango/pangocairo.h>
 #include "gdkcairo.h"
@@ -347,8 +347,6 @@ gdk_draw_line (GdkDrawable *drawable,
 {
   GdkSegment segment;
 
-  g_return_if_fail (drawable != NULL);
-  g_return_if_fail (gc != NULL);
   g_return_if_fail (GDK_IS_DRAWABLE (drawable));
   g_return_if_fail (GDK_IS_GC (gc));
 
@@ -467,22 +465,23 @@ gdk_draw_arc (GdkDrawable *drawable,
  *     necessary.
  * @points: an array of #GdkPoint structures specifying the points making 
  *     up the polygon.
- * @npoints: the number of points.
+ * @n_points: the number of points.
  * 
  * Draws an outlined or filled polygon.
  **/
 void
-gdk_draw_polygon (GdkDrawable *drawable,
-		  GdkGC       *gc,
-		  gboolean     filled,
-		  GdkPoint    *points,
-		  gint         npoints)
+gdk_draw_polygon (GdkDrawable    *drawable,
+		  GdkGC          *gc,
+		  gboolean        filled,
+		  const GdkPoint *points,
+		  gint            n_points)
 {
   g_return_if_fail (GDK_IS_DRAWABLE (drawable));
   g_return_if_fail (GDK_IS_GC (gc));
 
   GDK_DRAWABLE_GET_CLASS (drawable)->draw_polygon (drawable, gc, filled,
-                                                   points, npoints);
+                                                   (GdkPoint *) points,
+                                                   n_points);
 }
 
 /* gdk_draw_string
@@ -629,7 +628,7 @@ gdk_draw_drawable (GdkDrawable *drawable,
   gint composite_y_offset = 0;
 
   g_return_if_fail (GDK_IS_DRAWABLE (drawable));
-  g_return_if_fail (src != NULL);
+  g_return_if_fail (GDK_IS_DRAWABLE (src));
   g_return_if_fail (GDK_IS_GC (gc));
 
   if (width < 0 || height < 0)
@@ -692,7 +691,7 @@ gdk_draw_image (GdkDrawable *drawable,
 		gint         height)
 {
   g_return_if_fail (GDK_IS_DRAWABLE (drawable));
-  g_return_if_fail (image != NULL);
+  g_return_if_fail (GDK_IS_IMAGE (image));
   g_return_if_fail (GDK_IS_GC (gc));
 
   if (width == -1)
@@ -730,34 +729,45 @@ gdk_draw_image (GdkDrawable *drawable,
  *
  * The clip mask of @gc is ignored, but clip rectangles and clip regions work
  * fine.
- * 
+ *
+ * If GDK is built with the Sun mediaLib library, the gdk_draw_pixbuf
+ * function is accelerated using mediaLib, which provides hardware
+ * acceleration on Intel, AMD, and Sparc chipsets.  If desired, mediaLib
+ * support can be turned off by setting the GDK_DISABLE_MEDIALIB environment
+ * variable.
+ *
  * Since: 2.2
  **/
 void
 gdk_draw_pixbuf (GdkDrawable     *drawable,
-		  GdkGC           *gc,
-		  GdkPixbuf       *pixbuf,
-		  gint             src_x,
-		  gint             src_y,
-		  gint             dest_x,
-		  gint             dest_y,
-		  gint             width,
-		  gint             height,
-		  GdkRgbDither     dither,
-		  gint             x_dither,
-		  gint             y_dither)
+                 GdkGC           *gc,
+                 const GdkPixbuf *pixbuf,
+                 gint             src_x,
+                 gint             src_y,
+                 gint             dest_x,
+                 gint             dest_y,
+                 gint             width,
+                 gint             height,
+                 GdkRgbDither     dither,
+                 gint             x_dither,
+                 gint             y_dither)
 {
   g_return_if_fail (GDK_IS_DRAWABLE (drawable));
   g_return_if_fail (gc == NULL || GDK_IS_GC (gc));
   g_return_if_fail (GDK_IS_PIXBUF (pixbuf));
 
+  if (width == 0 || height == 0)
+    return;
+
   if (width == -1)
     width = gdk_pixbuf_get_width (pixbuf);
   if (height == -1)
     height = gdk_pixbuf_get_height (pixbuf);
 
-  GDK_DRAWABLE_GET_CLASS (drawable)->draw_pixbuf (drawable, gc, pixbuf,
-						  src_x, src_y, dest_x, dest_y, width, height,
+  GDK_DRAWABLE_GET_CLASS (drawable)->draw_pixbuf (drawable, gc,
+                                                  (GdkPixbuf *) pixbuf,
+						  src_x, src_y, dest_x, dest_y,
+                                                  width, height,
 						  dither, x_dither, y_dither);
 }
 
@@ -766,26 +776,27 @@ gdk_draw_pixbuf (GdkDrawable     *drawable,
  * @drawable: a #GdkDrawable (a #GdkWindow or a #GdkPixmap).
  * @gc: a #GdkGC.
  * @points: an array of #GdkPoint structures.
- * @npoints: the number of points to be drawn.
+ * @n_points: the number of points to be drawn.
  * 
  * Draws a number of points, using the foreground color and other 
  * attributes of the #GdkGC.
  **/
 void
-gdk_draw_points (GdkDrawable *drawable,
-		 GdkGC       *gc,
-		 GdkPoint    *points,
-		 gint         npoints)
+gdk_draw_points (GdkDrawable    *drawable,
+		 GdkGC          *gc,
+		 const GdkPoint *points,
+		 gint            n_points)
 {
   g_return_if_fail (GDK_IS_DRAWABLE (drawable));
-  g_return_if_fail ((points != NULL) && (npoints > 0));
+  g_return_if_fail ((points != NULL) && (n_points > 0));
   g_return_if_fail (GDK_IS_GC (gc));
-  g_return_if_fail (npoints >= 0);
+  g_return_if_fail (n_points >= 0);
 
-  if (npoints == 0)
+  if (n_points == 0)
     return;
 
-  GDK_DRAWABLE_GET_CLASS (drawable)->draw_points (drawable, gc, points, npoints);
+  GDK_DRAWABLE_GET_CLASS (drawable)->draw_points (drawable, gc,
+                                                  (GdkPoint *) points, n_points);
 }
 
 /**
@@ -794,27 +805,28 @@ gdk_draw_points (GdkDrawable *drawable,
  * @gc: a #GdkGC.
  * @segs: an array of #GdkSegment structures specifying the start and 
  *   end points of the lines to be drawn.
- * @nsegs: the number of line segments to draw, i.e. the size of the 
+ * @n_segs: the number of line segments to draw, i.e. the size of the 
  *   @segs array.
  * 
  * Draws a number of unconnected lines.
  **/
 void
-gdk_draw_segments (GdkDrawable *drawable,
-		   GdkGC       *gc,
-		   GdkSegment  *segs,
-		   gint         nsegs)
+gdk_draw_segments (GdkDrawable      *drawable,
+		   GdkGC            *gc,
+		   const GdkSegment *segs,
+		   gint              n_segs)
 {
   g_return_if_fail (GDK_IS_DRAWABLE (drawable));
 
-  if (nsegs == 0)
+  if (n_segs == 0)
     return;
 
   g_return_if_fail (segs != NULL);
   g_return_if_fail (GDK_IS_GC (gc));
-  g_return_if_fail (nsegs >= 0);
+  g_return_if_fail (n_segs >= 0);
 
-  GDK_DRAWABLE_GET_CLASS (drawable)->draw_segments (drawable, gc, segs, nsegs);
+  GDK_DRAWABLE_GET_CLASS (drawable)->draw_segments (drawable, gc,
+                                                    (GdkSegment *) segs, n_segs);
 }
 
 /**
@@ -822,7 +834,7 @@ gdk_draw_segments (GdkDrawable *drawable,
  * @drawable: a #GdkDrawable (a #GdkWindow or a #GdkPixmap).
  * @gc: a #GdkGC.
  * @points: an array of #GdkPoint structures specifying the endpoints of the
- * @npoints: the size of the @points array.
+ * @n_points: the size of the @points array.
  * 
  * Draws a series of lines connecting the given points.
  * The way in which joins between lines are draw is determined by the
@@ -830,30 +842,31 @@ gdk_draw_segments (GdkDrawable *drawable,
  * gdk_gc_set_line_attributes().
  **/
 void
-gdk_draw_lines (GdkDrawable *drawable,
-		GdkGC       *gc,
-		GdkPoint    *points,
-		gint         npoints)
+gdk_draw_lines (GdkDrawable    *drawable,
+		GdkGC          *gc,
+		const GdkPoint *points,
+		gint            n_points)
 {
   g_return_if_fail (GDK_IS_DRAWABLE (drawable));
   g_return_if_fail (points != NULL);
   g_return_if_fail (GDK_IS_GC (gc));
-  g_return_if_fail (npoints >= 0);
+  g_return_if_fail (n_points >= 0);
 
-  if (npoints == 0)
+  if (n_points == 0)
     return;
 
-  GDK_DRAWABLE_GET_CLASS (drawable)->draw_lines (drawable, gc, points, npoints);
+  GDK_DRAWABLE_GET_CLASS (drawable)->draw_lines (drawable, gc,
+                                                 (GdkPoint *) points, n_points);
 }
 
 static void
-real_draw_glyphs (GdkDrawable      *drawable,
-		  GdkGC	           *gc,
-		  PangoMatrix      *matrix,
-		  PangoFont        *font,
-		  gdouble           x,
-		  gdouble           y,
-		  PangoGlyphString *glyphs)
+real_draw_glyphs (GdkDrawable       *drawable,
+		  GdkGC	            *gc,
+		  const PangoMatrix *matrix,
+		  PangoFont         *font,
+		  gdouble            x,
+		  gdouble            y,
+		  PangoGlyphString  *glyphs)
 {
   cairo_t *cr;
 
@@ -938,13 +951,13 @@ gdk_draw_glyphs (GdkDrawable      *drawable,
  * Since: 2.6
  **/
 void
-gdk_draw_glyphs_transformed (GdkDrawable      *drawable,
-			     GdkGC	      *gc,
-			     PangoMatrix      *matrix,
-			     PangoFont        *font,
-			     gint              x,
-			     gint              y,
-			     PangoGlyphString *glyphs)
+gdk_draw_glyphs_transformed (GdkDrawable       *drawable,
+			     GdkGC	       *gc,
+			     const PangoMatrix *matrix,
+			     PangoFont         *font,
+			     gint               x,
+			     gint               y,
+			     PangoGlyphString  *glyphs)
 {
   g_return_if_fail (GDK_IS_DRAWABLE (drawable));
   g_return_if_fail (GDK_IS_GC (gc));
@@ -969,10 +982,10 @@ gdk_draw_glyphs_transformed (GdkDrawable      *drawable,
  * Since: 2.6
  **/
 void
-gdk_draw_trapezoids (GdkDrawable    *drawable,
-		     GdkGC	    *gc,
-		     GdkTrapezoid   *trapezoids,
-		     gint            n_trapezoids)
+gdk_draw_trapezoids (GdkDrawable        *drawable,
+		     GdkGC	        *gc,
+		     const GdkTrapezoid *trapezoids,
+		     gint                n_trapezoids)
 {
   cairo_t *cr;
   int i;
@@ -989,7 +1002,7 @@ gdk_draw_trapezoids (GdkDrawable    *drawable,
       cairo_move_to (cr, trapezoids[i].x11, trapezoids[i].y1);
       cairo_line_to (cr, trapezoids[i].x21, trapezoids[i].y1);
       cairo_line_to (cr, trapezoids[i].x22, trapezoids[i].y2);
-      cairo_line_to (cr, trapezoids[i].x21, trapezoids[i].y2);
+      cairo_line_to (cr, trapezoids[i].x12, trapezoids[i].y2);
       cairo_close_path (cr);
     }
 
@@ -1357,6 +1370,44 @@ composite_0888 (guchar      *src_buf,
     }
 }
 
+#ifdef USE_MEDIALIB
+static void
+composite_0888_medialib (guchar      *src_buf,
+			 gint         src_rowstride,
+			 guchar      *dest_buf,
+			 gint         dest_rowstride,
+			 GdkByteOrder dest_byte_order,
+			 gint         width,
+			 gint         height)
+{
+  guchar *src  = src_buf;
+  guchar *dest = dest_buf;
+
+  mlib_image img_src, img_dst;
+
+  mlib_ImageSetStruct (&img_dst,
+                       MLIB_BYTE,
+                       4,
+                       width,
+                       height,
+                       dest_rowstride,
+                       dest_buf);
+
+  mlib_ImageSetStruct (&img_src,
+                       MLIB_BYTE,
+                       4,
+                       width,
+                       height,
+                       src_rowstride,
+                       src_buf);
+
+  if (dest_byte_order == GDK_LSB_FIRST)
+      mlib_ImageBlendRGBA2BGRA (&img_dst, &img_src);
+  else
+      mlib_ImageBlendRGBA2ARGB (&img_dst, &img_src);
+}
+#endif
+
 static void
 composite_565 (guchar      *src_buf,
 	       gint         src_rowstride,
@@ -1452,7 +1503,7 @@ gdk_drawable_real_draw_pixbuf (GdkDrawable  *drawable,
   GdkRegion *clip;
   GdkRegion *drect;
   GdkRectangle tmp_rect;
-				       
+
   g_return_if_fail (GDK_IS_PIXBUF (pixbuf));
   g_return_if_fail (pixbuf->colorspace == GDK_COLORSPACE_RGB);
   g_return_if_fail (pixbuf->n_channels == 3 || pixbuf->n_channels == 4);
@@ -1554,7 +1605,16 @@ gdk_drawable_real_draw_pixbuf (GdkDrawable  *drawable,
 		   visual->red_mask   == 0xff0000 &&
 		   visual->green_mask == 0x00ff00 &&
 		   visual->blue_mask  == 0x0000ff)
-	    composite_func = composite_0888;
+	    {
+#ifdef USE_MEDIALIB
+	      if (_gdk_use_medialib ())
+	        composite_func = composite_0888_medialib;
+	      else
+	        composite_func = composite_0888;
+#else
+	      composite_func = composite_0888;
+#endif
+	    }
 	}
 
       /* We can't use our composite func if we are required to dither
