@@ -84,8 +84,7 @@ gail_image_cell_new (void)
   cell = GAIL_RENDERER_CELL(object);
 
   cell->renderer = gtk_cell_renderer_pixbuf_new ();
-  g_object_ref (cell->renderer);
-  gtk_object_sink (GTK_OBJECT (cell->renderer));
+  g_object_ref_sink (cell->renderer);
   return atk_object;
 }
 
